@@ -10,17 +10,21 @@ mod casing_style;
 mod error;
 mod parse;
 
+use attributes::FormatArg;
 use proc_macro2::TokenStream;
-use proc_macro_error::proc_macro_error;
+use proc_macro_error2::proc_macro_error;
 use quote::{quote, ToTokens, TokenStreamExt};
 use std::{collections::HashMap, str};
+use syn::visit_mut::VisitMut;
 use syn::{
-    parse_macro_input, token, Data, DataEnum, DataStruct, DeriveInput, Field, Fields, Ident, Index,
-    Type, Variant,
+    parse_macro_input, token, Data, DataEnum, DataStruct, DeriveInput, ExprPath, Field, Fields,
+    Ident, Index, PathSegment, Type, TypePath, Variant,
 };
 
-use attributes::{Attributes, FuncArg, StructAttributes};
-use error::Error;
+use crate::attributes::{FieldAttributes, TypeAttributes};
+use crate::error::Error;
+
+type FieldNameFn = fn(usize, &Field) -> TokenStream;
 
 #[proc_macro_derive(Tabled, attributes(tabled))]
 #[proc_macro_error]
@@ -31,21 +35,26 @@ pub fn tabled(input: proc_macro::TokenStream) -> proc_macro::TokenStream {
 }
 
 fn impl_tabled(ast: &DeriveInput) -> TokenStream {
-    let attrs = StructAttributes::parse(&ast.attrs)
+    let attrs = TypeAttributes::parse(&ast.attrs)
         .map_err(error::abort)
         .unwrap();
 
-    let length = get_tabled_length(ast, &attrs)
+    let tabled_trait_path = get_crate_name_expr(&attrs).map_err(error::abort).unwrap();
+
+    let length = get_tabled_length(ast, &attrs, &tabled_trait_path)
+        .map_err(error::abort)
+        .unwrap();
+    let info = collect_info(ast, &attrs, &tabled_trait_path)
         .map_err(error::abort)
         .unwrap();
-    let info = collect_info(ast, &attrs).map_err(error::abort).unwrap();
     let fields = info.values;
     let headers = info.headers;
 
     let name = &ast.ident;
     let (impl_generics, ty_generics, where_clause) = ast.generics.split_for_impl();
+
     let expanded = quote! {
-        impl #impl_generics ::tabled::Tabled for #name #ty_generics #where_clause {
+        impl #impl_generics #tabled_trait_path for #name #ty_generics #where_clause {
             const LENGTH: usize = #length;
 
             fn fields(&self) -> Vec<::std::borrow::Cow<'_, str>> {
@@ -61,34 +70,38 @@ fn impl_tabled(ast: &DeriveInput) -> TokenStream {
     expanded
 }
 
-fn get_tabled_length(ast: &DeriveInput, attrs: &StructAttributes) -> Result<TokenStream, Error> {
+fn get_tabled_length(
+    ast: &DeriveInput,
+    attrs: &TypeAttributes,
+    trait_path: &ExprPath,
+) -> Result<TokenStream, Error> {
     match &ast.data {
-        Data::Struct(data) => get_fields_length(&data.fields),
+        Data::Struct(data) => get_fields_length(&data.fields, trait_path),
         Data::Enum(data) => {
             if attrs.inline {
                 Ok(quote! { 1 })
             } else {
-                get_enum_length(data)
+                get_enum_length(data, trait_path)
             }
         }
         Data::Union(_) => Err(Error::message("Union type isn't supported")),
     }
 }
 
-fn get_fields_length(fields: &Fields) -> Result<TokenStream, Error> {
+fn get_fields_length(fields: &Fields, tabled_trait: &ExprPath) -> Result<TokenStream, Error> {
     let size_components = fields
         .iter()
         .map(|field| {
-            let attributes = Attributes::parse(&field.attrs)?;
+            let attributes = FieldAttributes::parse(&field.attrs)?;
             Ok((field, attributes))
         })
         .collect::<Result<Vec<_>, Error>>()?
         .into_iter()
-        .filter(|(_, attr)| !attr.is_ignored())
+        .filter(|(_, attr)| !attr.is_ignored)
         .map(|(field, attr)| {
             if attr.inline {
                 let field_type = &field.ty;
-                quote!({<#field_type as Tabled>::LENGTH})
+                quote!({<#field_type as #tabled_trait>::LENGTH})
             } else {
                 quote!({ 1 })
             }
@@ -97,20 +110,20 @@ fn get_fields_length(fields: &Fields) -> Result<TokenStream, Error> {
     let size_components = std::iter::once(quote!(0)).chain(size_components);
 
     let mut stream = TokenStream::new();
-    stream.append_separated(size_components, syn::token::Add::default());
+    stream.append_separated(size_components, syn::token::Plus::default());
 
     Ok(stream)
 }
 
-fn get_enum_length(enum_ast: &DataEnum) -> Result<TokenStream, Error> {
-    let variant_sizes = get_enum_variant_length(enum_ast);
+fn get_enum_length(enum_ast: &DataEnum, trait_path: &ExprPath) -> Result<TokenStream, Error> {
+    let variant_sizes = get_enum_variant_length(enum_ast, trait_path);
 
     let mut stream = TokenStream::new();
     for (i, size) in variant_sizes.enumerate() {
         let size = size?;
 
         if i != 0 {
-            stream.append_all(syn::token::Add::default().into_token_stream());
+            stream.append_all(syn::token::Plus::default().into_token_stream());
         }
 
         stream.append_all(size);
@@ -119,38 +132,47 @@ fn get_enum_length(enum_ast: &DataEnum) -> Result<TokenStream, Error> {
     Ok(stream)
 }
 
-fn get_enum_variant_length(
-    enum_ast: &DataEnum,
-) -> impl Iterator<Item = Result<TokenStream, Error>> + '_ {
+fn get_enum_variant_length<'a>(
+    enum_ast: &'a DataEnum,
+    trait_path: &'a ExprPath,
+) -> impl Iterator<Item = Result<TokenStream, Error>> + 'a {
     enum_ast
         .variants
         .iter()
         .map(|variant| -> Result<_, Error> {
-            let attributes = Attributes::parse(&variant.attrs)?;
+            let attributes = FieldAttributes::parse(&variant.attrs)?;
             Ok((variant, attributes))
         })
-        .filter(|result| result.is_err() || matches!(result, Ok((_, attr)) if !attr.is_ignored()))
-        .map(|result| {
+        .filter(|result| result.is_err() || matches!(result, Ok((_, attr)) if !attr.is_ignored))
+        .map(move |result| {
             let (variant, attr) = result?;
 
             if attr.inline {
-                get_fields_length(&variant.fields)
+                get_fields_length(&variant.fields, trait_path)
             } else {
                 Ok(quote!(1))
             }
         })
 }
 
-fn collect_info(ast: &DeriveInput, attrs: &StructAttributes) -> Result<Impl, Error> {
+fn collect_info(
+    ast: &DeriveInput,
+    attrs: &TypeAttributes,
+    trait_path: &ExprPath,
+) -> Result<Impl, Error> {
     match &ast.data {
-        Data::Struct(data) => collect_info_struct(data, attrs),
-        Data::Enum(data) => collect_info_enum(data, attrs, &ast.ident),
+        Data::Struct(data) => collect_info_struct(data, attrs, trait_path),
+        Data::Enum(data) => collect_info_enum(data, attrs, &ast.ident, trait_path),
         Data::Union(_) => Err(Error::message("Union type isn't supported")),
     }
 }
 
-fn collect_info_struct(ast: &DataStruct, attrs: &StructAttributes) -> Result<Impl, Error> {
-    info_from_fields(&ast.fields, attrs, field_var_name, "")
+fn collect_info_struct(
+    ast: &DataStruct,
+    attrs: &TypeAttributes,
+    trait_path: &ExprPath,
+) -> Result<Impl, Error> {
+    info_from_fields(&ast.fields, attrs, struct_field_name, "", trait_path)
 }
 
 // todo: refactoring. instead of using a lambda + prefix
@@ -158,17 +180,18 @@ fn collect_info_struct(ast: &DataStruct, attrs: &StructAttributes) -> Result<Imp
 // So the called would prefix it on its own
 fn info_from_fields(
     fields: &Fields,
-    attrs: &StructAttributes,
-    field_name: impl Fn(usize, &Field) -> TokenStream,
+    attrs: &TypeAttributes,
+    field_name: FieldNameFn,
     header_prefix: &str,
+    trait_path: &ExprPath,
 ) -> Result<Impl, Error> {
     let count_fields = fields.len();
 
-    let fields = fields
+    let attributes = fields
         .into_iter()
         .enumerate()
         .map(|(i, field)| -> Result<_, Error> {
-            let mut attributes = Attributes::parse(&field.attrs)?;
+            let mut attributes = FieldAttributes::parse(&field.attrs)?;
             merge_attributes(&mut attributes, attrs);
 
             Ok((i, field, attributes))
@@ -179,9 +202,9 @@ fn info_from_fields(
     let mut reorder = HashMap::new();
 
     let mut skipped = 0;
-    for result in fields {
+    for result in attributes {
         let (i, field, attributes) = result?;
-        if attributes.is_ignored() {
+        if attributes.is_ignored {
             skipped += 1;
             continue;
         }
@@ -196,11 +219,18 @@ fn info_from_fields(
             reorder.insert(order, i - skipped);
         }
 
-        let header = field_headers(field, i, &attributes, header_prefix);
+        let header = field_headers(field, i, &attributes, header_prefix, trait_path);
         headers.push(header);
 
-        let field_name = field_name(i, field);
-        let value = get_field_fields(&field_name, &attributes);
+        let field_name_result = field_name(i, field);
+        let value = get_field_fields(
+            &field_name_result,
+            &field.ty,
+            &attributes,
+            fields,
+            field_name,
+            attrs,
+        );
         values.push(value);
     }
 
@@ -255,15 +285,16 @@ fn reorder_fields<T: Clone>(order: &HashMap<usize, usize>, elements: &[T]) -> Ve
 fn field_headers(
     field: &Field,
     index: usize,
-    attributes: &Attributes,
+    attributes: &FieldAttributes,
     prefix: &str,
+    trait_path: &ExprPath,
 ) -> TokenStream {
     if attributes.inline {
         let prefix = attributes
             .inline_prefix
             .as_ref()
             .map_or_else(|| "", |s| s.as_str());
-        return get_type_headers(&field.ty, prefix, "");
+        return get_type_headers(&field.ty, prefix, "", trait_path);
     }
 
     let header_name = field_header_name(field, attributes, index);
@@ -277,8 +308,9 @@ fn field_headers(
 
 fn collect_info_enum(
     ast: &DataEnum,
-    attrs: &StructAttributes,
+    attrs: &TypeAttributes,
     name: &Ident,
+    trait_path: &ExprPath,
 ) -> Result<Impl, Error> {
     match &attrs.inline {
         true => {
@@ -289,35 +321,39 @@ fn collect_info_enum(
 
             collect_info_enum_inlined(ast, attrs, enum_name)
         }
-        false => _collect_info_enum(ast, attrs),
+        false => _collect_info_enum(ast, attrs, trait_path),
     }
 }
 
-fn _collect_info_enum(ast: &DataEnum, attrs: &StructAttributes) -> Result<Impl, Error> {
+fn _collect_info_enum(
+    ast: &DataEnum,
+    attrs: &TypeAttributes,
+    trait_path: &ExprPath,
+) -> Result<Impl, Error> {
     // reorder variants according to order (if set)
     let orderedvariants = reodered_variants(ast)?;
 
     let mut headers_list = Vec::new();
     let mut variants = Vec::new();
     for v in orderedvariants {
-        let mut attributes = Attributes::parse(&v.attrs)?;
+        let mut attributes = FieldAttributes::parse(&v.attrs)?;
         merge_attributes(&mut attributes, attrs);
-        if attributes.is_ignored() {
+        if attributes.is_ignored {
             continue;
         }
 
-        let info = info_from_variant(v, &attributes, attrs)?;
+        let info = info_from_variant(v, &attributes, attrs, trait_path)?;
         variants.push((v, info.values));
         headers_list.push(info.headers);
     }
 
-    let variant_sizes = get_enum_variant_length(ast)
+    let variant_sizes = get_enum_variant_length(ast, trait_path)
         .collect::<Result<Vec<_>, Error>>()?
         .into_iter();
-    let values = values_for_enum(variant_sizes, &variants);
+    let values = values_for_enum(variant_sizes, &variants, trait_path);
 
     let headers = quote! {
-        vec![
+        [
             #(#headers_list,)*
         ]
         .concat()
@@ -328,7 +364,7 @@ fn _collect_info_enum(ast: &DataEnum, attrs: &StructAttributes) -> Result<Impl,
 
 fn collect_info_enum_inlined(
     ast: &DataEnum,
-    attrs: &StructAttributes,
+    attrs: &TypeAttributes,
     enum_name: String,
 ) -> Result<Impl, Error> {
     let orderedvariants = reodered_variants(ast)?;
@@ -336,10 +372,10 @@ fn collect_info_enum_inlined(
     let mut variants = Vec::new();
     let mut names = Vec::new();
     for variant in orderedvariants {
-        let mut attributes = Attributes::parse(&variant.attrs)?;
+        let mut attributes = FieldAttributes::parse(&variant.attrs)?;
         merge_attributes(&mut attributes, attrs);
         let mut name = String::new();
-        if !attributes.is_ignored() {
+        if !attributes.is_ignored {
             name = variant_name(variant, &attributes);
         }
 
@@ -360,33 +396,45 @@ fn collect_info_enum_inlined(
 
 fn info_from_variant(
     variant: &Variant,
-    attr: &Attributes,
-    attrs: &StructAttributes,
+    attr: &FieldAttributes,
+    attrs: &TypeAttributes,
+    trait_path: &ExprPath,
 ) -> Result<Impl, Error> {
     if attr.inline {
         let prefix = attr
             .inline_prefix
             .as_ref()
             .map_or_else(|| "", |s| s.as_str());
-        return info_from_fields(&variant.fields, attrs, variant_var_name, prefix);
+        return info_from_fields(
+            &variant.fields,
+            attrs,
+            variant_field_name,
+            prefix,
+            trait_path,
+        );
     }
 
     let variant_name = variant_name(variant, attr);
     let value = if let Some(func) = &attr.display_with {
         let args = match &attr.display_with_args {
-            None => None,
-            Some(args) => match args.is_empty() {
-                true => None,
-                false => {
-                    let args = args.iter().map(fnarg_tokens).collect::<Vec<_>>();
-                    Some(quote!( #(#args,)* ))
-                }
-            },
+            Some(args) => {
+                args_to_tokens_with(&Fields::Unit, &quote!(self), struct_field_name, args)
+            }
+            None => quote!(&self),
         };
 
+        let result = use_function(&args, func);
+
+        quote! { ::std::borrow::Cow::from(#result) }
+    } else if let Some(fmt) = &attr.format {
+        let args = attr
+            .format_with_args
+            .as_ref()
+            .and_then(|args| args_to_tokens(&Fields::Unit, struct_field_name, args));
+
         let call = match args {
-            Some(args) => use_function(&args, func),
-            None => use_function_no_args(func),
+            Some(args) => use_format(fmt, &args),
+            None => use_format_no_args(fmt),
         };
 
         quote! { ::std::borrow::Cow::from(#call) }
@@ -408,12 +456,17 @@ struct Impl {
     values: TokenStream,
 }
 
-fn get_type_headers(field_type: &Type, inline_prefix: &str, prefix: &str) -> TokenStream {
+fn get_type_headers(
+    field_type: &Type,
+    inline_prefix: &str,
+    prefix: &str,
+    tabled_trait: &ExprPath,
+) -> TokenStream {
     if prefix.is_empty() && inline_prefix.is_empty() {
-        quote! { <#field_type as Tabled>::headers() }
+        quote! { <#field_type as #tabled_trait>::headers() }
     } else {
         quote! {
-            <#field_type as Tabled>::headers().into_iter()
+            <#field_type as #tabled_trait>::headers().into_iter()
                 .map(|header| {
                     let header = format!("{}{}{}", #prefix, #inline_prefix, header);
                     ::std::borrow::Cow::Owned(header)
@@ -423,34 +476,126 @@ fn get_type_headers(field_type: &Type, inline_prefix: &str, prefix: &str) -> Tok
     }
 }
 
-fn get_field_fields(field: &TokenStream, attr: &Attributes) -> TokenStream {
+fn get_field_fields(
+    field: &TokenStream,
+    field_type: &Type,
+    attr: &FieldAttributes,
+    fields: &Fields,
+    field_name: FieldNameFn,
+    type_attrs: &TypeAttributes,
+) -> TokenStream {
     if attr.inline {
         return quote! { #field.fields() };
     }
 
     if let Some(func) = &attr.display_with {
         let args = match &attr.display_with_args {
-            None => Some(quote!(&#field)),
-            Some(args) => match args.is_empty() {
-                true => None,
-                false => {
-                    let args = args.iter().map(fnarg_tokens).collect::<Vec<_>>();
-                    Some(quote!( #(#args,)* ))
-                }
-            },
+            Some(args) => args_to_tokens_with(fields, field, field_name, args),
+            None => quote!(&#field),
         };
 
+        let result = use_function(&args, func);
+
+        return quote!(vec![::std::borrow::Cow::from(#result)]);
+    } else if let Some(fmt) = &attr.format {
+        let args = attr
+            .format_with_args
+            .as_ref()
+            .and_then(|args| args_to_tokens(fields, field_name, args));
+
         let call = match args {
-            Some(args) => use_function(&args, func),
-            None => use_function_no_args(func),
+            Some(args) => use_format(fmt, &args),
+            None => use_format_with_one_arg(fmt, field),
         };
 
-        return quote!(vec![::std::borrow::Cow::from(#call)]);
+        return quote!(vec![::std::borrow::Cow::Owned(#call)]);
+    }
+
+    if let Some(i) = find_display_type(field_type, &type_attrs.display_types) {
+        let (_, func, args) = &type_attrs.display_types[i];
+        let args = args_to_tokens_with(fields, field, field_name, args);
+        let func = use_function(&args, func);
+
+        return quote!(vec![::std::borrow::Cow::from(#func)]);
     }
 
     quote!(vec![::std::borrow::Cow::Owned(format!("{}", #field))])
 }
 
+fn args_to_tokens(
+    fields: &Fields,
+    field_name: fn(usize, &Field) -> TokenStream,
+    args: &[FormatArg],
+) -> Option<TokenStream> {
+    if args.is_empty() {
+        return None;
+    }
+
+    let args = args
+        .iter()
+        .map(|arg| fnarg_tokens(arg, fields, field_name))
+        .collect::<Vec<_>>();
+    Some(quote!( #(#args,)* ))
+}
+
+fn args_to_tokens_with(
+    fields: &Fields,
+    field: &TokenStream,
+    field_name: fn(usize, &Field) -> TokenStream,
+    args: &[FormatArg],
+) -> TokenStream {
+    if args.is_empty() {
+        return quote!(&#field);
+    }
+
+    let mut out = vec![quote!(&#field)];
+    for arg in args {
+        let arg = fnarg_tokens(arg, fields, field_name);
+        out.push(arg);
+    }
+
+    quote!( #(#out,)* )
+}
+
+fn find_display_type(ty: &Type, types: &[(TypePath, String, Vec<FormatArg>)]) -> Option<usize> {
+    let p: &TypePath = match ty {
+        Type::Path(path) => path,
+        _ => return None,
+    };
+
+    // NOTICE:
+    // We do iteration in a back order to satisfy a later set argument first.
+    //
+    // TODO: Maybe we shall change the data structure for it rather then doing a reverse iteration?
+    // I am just saying it's dirty a little.
+    let args = types.iter().enumerate().rev();
+    for (i, (arg, _, _)) in args {
+        if arg.path == p.path {
+            return Some(i);
+        }
+
+        // NOTICE:
+        // There's a specical case where we wanna match a type without a generic,
+        // e.g. 'Option' with which we wanna match all 'Option's.
+        //
+        // Because in the scope we can only have 1 type name, it's considered to be good,
+        // and nothing must be broken.
+        let arg_segment = arg.path.segments.last();
+        let type_segment = p.path.segments.last();
+        if let Some(arg) = arg_segment {
+            if arg.arguments.is_empty() {
+                if let Some(p) = type_segment {
+                    if p.ident == arg.ident {
+                        return Some(i);
+                    }
+                }
+            }
+        }
+    }
+
+    None
+}
+
 fn use_function(args: &TokenStream, function: &str) -> TokenStream {
     let path: syn::Result<syn::ExprPath> = syn::parse_str(function);
     match path {
@@ -464,20 +609,19 @@ fn use_function(args: &TokenStream, function: &str) -> TokenStream {
     }
 }
 
-fn use_function_no_args(function: &str) -> TokenStream {
-    let path: syn::Result<syn::ExprPath> = syn::parse_str(function);
-    match path {
-        Ok(path) => {
-            quote! { #path() }
-        }
-        Err(_) => {
-            let function = Ident::new(function, proc_macro2::Span::call_site());
-            quote! { #function() }
-        }
-    }
+fn use_format(custom_format: &str, args: &TokenStream) -> TokenStream {
+    quote! { format!(#custom_format, #args) }
 }
 
-fn field_var_name(index: usize, field: &Field) -> TokenStream {
+fn use_format_with_one_arg(custom_format: &str, field: &TokenStream) -> TokenStream {
+    quote! { format!(#custom_format, #field) }
+}
+
+fn use_format_no_args(custom_format: &str) -> TokenStream {
+    quote! { format!(#custom_format) }
+}
+
+fn struct_field_name(index: usize, field: &Field) -> TokenStream {
     let f = field.ident.as_ref().map_or_else(
         || Index::from(index).to_token_stream(),
         quote::ToTokens::to_token_stream,
@@ -485,7 +629,7 @@ fn field_var_name(index: usize, field: &Field) -> TokenStream {
     quote!(self.#f)
 }
 
-fn variant_var_name(index: usize, field: &Field) -> TokenStream {
+fn variant_field_name(index: usize, field: &Field) -> TokenStream {
     match &field.ident {
         Some(indent) => indent.to_token_stream(),
         None => Ident::new(
@@ -499,6 +643,7 @@ fn variant_var_name(index: usize, field: &Field) -> TokenStream {
 fn values_for_enum(
     variant_sizes: impl Iterator<Item = TokenStream>,
     variants: &[(&Variant, TokenStream)],
+    tabled_trait: &ExprPath,
 ) -> TokenStream {
     let branches = variants.iter().map(|(variant, _)| match_variant(variant));
 
@@ -536,7 +681,7 @@ fn values_for_enum(
             offsets[i] += offsets[i-1]
         }
 
-        let size = <Self as Tabled>::LENGTH;
+        let size = <Self as #tabled_trait>::LENGTH;
         let mut out_vec = vec![::std::borrow::Cow::Borrowed(""); size];
 
         #[allow(unused_variables)]
@@ -555,7 +700,7 @@ fn variant_idents(v: &Variant) -> Vec<TokenStream> {
         .enumerate()
         // we intentionally not ignore these fields to be able to build a pattern correctly
         // .filter(|(_, field)| !Attr::parse(&field.attrs).is_ignored())
-        .map(|(index, field)| variant_var_name(index, field))
+        .map(|(index, field)| variant_field_name(index, field))
         .collect::<Vec<_>>()
 }
 
@@ -582,7 +727,7 @@ fn match_variant(v: &Variant) -> TokenStream {
     token
 }
 
-fn variant_name(variant: &Variant, attributes: &Attributes) -> String {
+fn variant_name(variant: &Variant, attributes: &FieldAttributes) -> String {
     attributes
         .rename
         .clone()
@@ -595,7 +740,7 @@ fn variant_name(variant: &Variant, attributes: &Attributes) -> String {
         .unwrap_or_else(|| variant.ident.to_string())
 }
 
-fn field_header_name(f: &Field, attr: &Attributes, index: usize) -> String {
+fn field_header_name(f: &Field, attr: &FieldAttributes, index: usize) -> String {
     if let Some(name) = &attr.rename {
         return name.to_string();
     }
@@ -612,29 +757,97 @@ fn field_header_name(f: &Field, attr: &Attributes, index: usize) -> String {
     }
 }
 
-fn merge_attributes(attr: &mut Attributes, global_attr: &StructAttributes) {
+fn merge_attributes(attr: &mut FieldAttributes, global_attr: &TypeAttributes) {
     if attr.rename_all.is_none() {
         attr.rename_all = global_attr.rename_all;
     }
 }
 
-fn fnarg_tokens(arg: &FuncArg) -> TokenStream {
-    match arg {
-        FuncArg::SelfRef => quote! { &self },
-        FuncArg::Byte(val) => quote! { #val },
-        FuncArg::Char(val) => quote! { #val },
-        FuncArg::Bool(val) => quote! { #val },
-        FuncArg::Uint(val) => quote! { #val },
-        FuncArg::Int(val) => quote! { #val },
-        FuncArg::Float(val) => quote! { #val },
-        FuncArg::String(val) => quote! { #val },
-        FuncArg::Bytes(val) => {
-            let val = syn::LitByteStr::new(val, proc_macro2::Span::call_site());
-            quote! { #val }
+// to resolve invocation issues withing a macros calls
+// we need such a workround
+struct ExprSelfReplace<'a>(Option<(&'a Fields, FieldNameFn)>);
+
+impl syn::visit_mut::VisitMut for ExprSelfReplace<'_> {
+    fn visit_expr_mut(&mut self, node: &mut syn::Expr) {
+        match &node {
+            syn::Expr::Path(path) => {
+                let indent = path.path.get_ident();
+                if let Some(indent) = indent {
+                    if indent == "self" {
+                        *node = syn::parse_quote! { (&self) };
+                        return;
+                    }
+                }
+            }
+            syn::Expr::Field(field) => {
+                // treating a enum variant structs which we can't reference by 'self'
+                if let Some((fields, field_name)) = &self.0 {
+                    // check that it's plain self reference
+
+                    if let syn::Expr::Path(path) = field.base.as_ref() {
+                        let indent = path.path.get_ident();
+                        if let Some(indent) = indent {
+                            if indent != "self" {
+                                return;
+                            }
+                        }
+                    }
+
+                    let used_field = {
+                        match &field.member {
+                            syn::Member::Named(ident) => ident.to_string(),
+                            syn::Member::Unnamed(index) => index.index.to_string(),
+                        }
+                    };
+
+                    // We find the corresponding field in the local object fields instead of using self,
+                    // which would be a higher level object. This is for nested structures.
+
+                    for (i, field) in fields.iter().enumerate() {
+                        let field_name_result = (field_name)(i, field);
+                        let field_name = field
+                            .ident
+                            .as_ref()
+                            .map_or_else(|| i.to_string(), |i| i.to_string());
+                        if field_name == used_field {
+                            *node = syn::parse_quote! { #field_name_result };
+                            return;
+                        }
+                    }
+                }
+            }
+            syn::Expr::Macro(_) => {
+                // NOTE: Can we parse inners of Macros?
+                //
+                // A next example will fail on `self.f1` usage
+                // with such error 'expected value, found module `self`'
+                //
+                // ```
+                // some_macro! {
+                //     struct Something {
+                //         #[tabled(display("_", format!("", self.f1)))]
+                //         field: Option<sstr>,
+                //         f1: usize,
+                //     }
+                // }
+                // ```
+            }
+            _ => (),
         }
+
+        // Delegate to the default impl to visit nested expressions.
+        syn::visit_mut::visit_expr_mut(self, node);
     }
 }
 
+fn fnarg_tokens(arg: &FormatArg, fields: &Fields, field_name: FieldNameFn) -> TokenStream {
+    let mut exp = arg.expr.clone();
+
+    ExprSelfReplace(Some((fields, field_name))).visit_expr_mut(&mut exp);
+
+    quote!(#exp)
+}
+
 fn reodered_variants(ast: &DataEnum) -> Result<Vec<&Variant>, Error> {
     let mut reorder = HashMap::new();
     let mut skip = 0;
@@ -642,7 +855,7 @@ fn reodered_variants(ast: &DataEnum) -> Result<Vec<&Variant>, Error> {
     for (i, attr) in ast
         .variants
         .iter()
-        .map(|v| Attributes::parse(&v.attrs).unwrap_or_default())
+        .map(|v| FieldAttributes::parse(&v.attrs).unwrap_or_default())
         .enumerate()
     {
         if attr.is_ignored {
@@ -668,3 +881,24 @@ fn reodered_variants(ast: &DataEnum) -> Result<Vec<&Variant>, Error> {
 
     Ok(orderedvariants)
 }
+
+fn get_crate_name_expr(attrs: &TypeAttributes) -> Result<ExprPath, Error> {
+    let crate_name = attrs
+        .crate_name
+        .clone()
+        .unwrap_or_else(|| String::from("::tabled"));
+    let crate_name = parse_crate_name(&crate_name)?;
+    Ok(create_tabled_trait_path(crate_name))
+}
+
+fn parse_crate_name(name: &str) -> Result<ExprPath, Error> {
+    syn::parse_str(name).map_err(|_| Error::message("unexpected crate attribute type"))
+}
+
+fn create_tabled_trait_path(mut p: ExprPath) -> ExprPath {
+    p.path.segments.push(PathSegment {
+        ident: Ident::new("Tabled", proc_macro2::Span::call_site()),
+        arguments: syn::PathArguments::None,
+    });
+    p
+}
